#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <io.h>
#include <conio.h>
#include <windows.h>
#include <winioctl.h>

#define BUFSIZE MAX_PATH        
#define FILESYSNAMEBUFSIZE MAX_PATH

/*
display BIOS Parameter Block(BPB)

Command prompt
Windows NT 4.0 or later

bpb [drive]

ex.
bpb e:
*/

typedef struct {
  int PartitionType;
  char *type;
}  PARTITION_LIST;

typedef struct {
  UINT DriveType;
  char *type;
}  DRIVE_LIST;

int lasterror = 0;
FILE *errout = stderr;
char *pgname = "bpb";

PARTITION_LIST PartitionList[] = {
{PARTITION_ENTRY_UNUSED ,"An unused entry."},
{PARTITION_EXTENDED ,"Extended partition entry."},
{PARTITION_FAT_12 ,"Partition with 12-bit FAT entries."},
{PARTITION_FAT_16 ,"Partition with 16-bit FAT entries."},
{PARTITION_FAT32 ,"FAT32 partition."},
{PARTITION_FAT32_XINT13 ,"FAT32 partition that uses extended int13 services."},
{PARTITION_HUGE ,"MS-DOS V4 huge partition."},
{PARTITION_IFS ,"IFS partition."},
{PARTITION_LDM ,"Logical disk manager (LDM) partition."},
{PARTITION_NTFT ,"NTFT partition."},
/*{PARTITION_OS2BOOTMGR ,"OS/2 boot manager partition."},*/
{PARTITION_PREP ,"PowerPC Reference Platform (PreP) partition."},
{PARTITION_UNIX ,"UNIX partition."},
{PARTITION_XENIX_1 ,"Xenix type 1 partition."},
{PARTITION_XENIX_2 ,"Xenix type 2 partition."},
{PARTITION_XINT13 ,"Partition that uses extended int13 services."},
{PARTITION_XINT13_EXTENDED ,"Extended partition that uses extended int13 services."},
{VALID_NTFT ,"The high bit of the partition type code indicates that a partition is part of an NTFT mirror or striped array. "},
{-1 ,"UNKNOWN"}
};

DRIVE_LIST DriveList[] = {
{DRIVE_UNKNOWN,"cannot be determined."},
{DRIVE_NO_ROOT_DIR,"invalid. For example, no volume is mounted at the path."},
{DRIVE_REMOVABLE,"removable drive"},
{DRIVE_FIXED,"fixed drive"},
{DRIVE_REMOTE,"remote (network) drive"},
{DRIVE_CDROM,"CD-ROM drive"},
{DRIVE_RAMDISK,"RAM disk"},
{(UINT)-1 ,"UNKNOWN"}
};

void lasterr(char *fmt, ...)
{
    LPVOID lpMsgBuf;
    va_list argp;

    lasterror = GetLastError();
    fprintf(errout,"%s: error %d ",pgname,lasterror);
    if ( fmt != NULL && *fmt != 0 ) {
        va_start(argp, fmt);
        vfprintf(errout,fmt,argp);
        va_end(argp);
    }
    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        lasterror,
        MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0,
        NULL
    );
    fprintf(errout,"%s",lpMsgBuf);
    LocalFree(lpMsgBuf);
}

char *partitionstring(int media)
{
    PARTITION_LIST *p = PartitionList;

    while ( p->PartitionType != -1 ) {
        if ( media == p->PartitionType ) {
            return p->type;
        }
        p++;
    }
    return p->type;
}

char *drivetypestring(UINT media)
{
    DRIVE_LIST *p = DriveList;

    while ( p->DriveType != (UINT)-1 ) {
        if ( media == p->DriveType ) {
            return p->type;
        }
        p++;
    }
    return p->type;
}

void ntfs(unsigned char *bpb)
{
    printf("00h-02h Jump Instruction: 0x%02x 0x%02x 0x%02x\n",bpb[0],bpb[1],bpb[2]);
    printf("03h-0ah OEM ID: \"%.8s\"\n",bpb + 0x3);
    printf("BIOS Parameter Block\n");
    printf("0bh-0ch Bytes Per Sector: 0x%x %d\n",*(short *)(bpb + 0x0b),*(short *)(bpb + 0x0b));
    printf("0dh     Sectors Per Cluster: 0x%x %d\n",bpb[0x0d],bpb[0x0d]);
    printf("0eh     Reserved Sectors: 0x%x %d\n",*(short *)(bpb + 0x0e),*(short *)(bpb + 0x0e));
    printf("10h-12h always 0: 0x%02x 0x%02x 0x%02x\n",bpb[0x10],bpb[0x11],bpb[0x12]);
    printf("13h-14h not used by NTFS: 0x%x %d\n",*(short *)(bpb + 0x13),*(short *)(bpb + 0x13));
    printf("15h     Media Descriptor: 0x%x %d\n",bpb[0x15],bpb[0x15]);
    printf("16h-17h always 0: 0x%x %d\n",*(short *)(bpb + 0x16),*(short *)(bpb + 0x16));
    printf("18h-19h Sectors Per Track: 0x%x %u\n",*(unsigned short *)(bpb + 0x18),*(unsigned short *)(bpb + 0x18));
    printf("1ah-1bh Number Of Heads: 0x%x %d\n",*(short *)(bpb + 0x1a),*(short *)(bpb + 0x1a));
    printf("1ch-1fh Hidden Sectors: 0x%x %u\n",*(unsigned *)(bpb + 0x1c),*(unsigned *)(bpb + 0x1c));
    printf("20h-23h not used by NTFS: 0x%x %u\n",*(unsigned *)(bpb + 0x20),*(unsigned *)(bpb + 0x20));
    printf("24h-27h not used by NTFS: 0x%x %u\n",*(unsigned *)(bpb + 0x24),*(unsigned *)(bpb + 0x24));
    printf("28h-2fh Total Sectors: 0x%I64x %I64d\n",*((LONGLONG *)(bpb + 0x28)+0),*((LONGLONG *)(bpb + 0x28)+0));
    printf("30h-37h Logical Cluster Number for the file $MFT: 0x%I64x %I64d\n",*((LONGLONG *)(bpb + 0x30)+0),*((LONGLONG *)(bpb + 0x37)+0));
    printf("38h-3fh Logical Cluster Number for the file $MFTMirr: 0x%I64x %I64d\n",*((LONGLONG *)(bpb + 0x38)+0),*((LONGLONG *)(bpb + 0x38)+0));
    printf("40h-43h Clusters Per File Record Segment: 0x%x %u\n",*(unsigned *)(bpb + 0x40),*(unsigned *)(bpb + 0x40));
    printf("44h-47h Clusters Per Index Block: 0x%x %u\n",*(unsigned *)(bpb + 0x44),*(unsigned *)(bpb + 0x44));
    printf("48h-4fh Volume Serial Number: 0x%I64x %I64d\n",*((LONGLONG *)(bpb + 0x48)+0),*((LONGLONG *)(bpb + 0x48)+0));
    printf("50h-53h Checksum: 0x%x %u\n",*(unsigned *)(bpb + 0x50),*(unsigned *)(bpb + 0x50));
}

void fat32(unsigned char *bpb)
{
    printf("00h-02h Jump Instruction: 0x%02x 0x%02x 0x%02x\n",bpb[0],bpb[1],bpb[2]);
    printf("03h-0ah OEM ID: \"%.8s\"\n",bpb + 0x3);
    printf("BIOS Parameter Block\n");
    printf("0bh-0ch Bytes Per Sector: 0x%x %d\n",*(short *)(bpb + 0x0b),*(short *)(bpb + 0x0b));
    printf("0dh     Sectors Per Cluster: 0x%x %d\n",bpb[0x0d],bpb[0x0d]);
    printf("0eh-0fh Reserved Sectors: 0x%x %d\n",*(short *)(bpb + 0x0e),*(short *)(bpb + 0x0e));
    printf("10h     Number of FATs: 0x%x %d\n",bpb[0x10],bpb[0x10]);
    printf("11h-12h Root Entries: 0x%x %u\n",*(unsigned short *)(bpb + 0x11),*(unsigned short *)(bpb + 0x11));
    printf("13h-14h Small Sectors: 0x%x %d\n",*(short *)(bpb + 0x13),*(short *)(bpb + 0x13));
    printf("15h     Media Descriptor: 0x%x %d\n",bpb[0x15],bpb[0x15]);
    printf("16h-17h Sectors Per FAT: 0x%x %d\n",*(short *)(bpb + 0x16),*(short *)(bpb + 0x16));
    printf("18h-19h Sectors Per Track: 0x%x %u\n",*(unsigned short *)(bpb + 0x18),*(unsigned short *)(bpb + 0x18));
    printf("1ah-1bh Number Of Heads: 0x%x %d\n",*(short *)(bpb + 0x1a),*(short *)(bpb + 0x1a));
    printf("1ch-1fh Hidden Sectors: 0x%x %u\n",*(unsigned *)(bpb + 0x1c),*(unsigned *)(bpb + 0x1c));
    printf("20h-23h Large Sectors: 0x%x %u\n",*(unsigned *)(bpb + 0x20),*(unsigned *)(bpb + 0x20));
    printf("24h-27h Sectors Per FAT: 0x%x\n",*(unsigned *)(bpb + 0x24),*(unsigned *)(bpb + 0x24));
    printf("28h-29h Extended Flags: 0x%x (Bit7:%d Bit0-3:%d)\n",*(unsigned *)(bpb + 0x28),(*(unsigned *)(bpb + 0x28)/128) & 1,*(unsigned *)(bpb + 0x28) & 15);
    printf("2ah-2bh File System Version: 0x%04x (%d.%d)\n",*(short *)(bpb + 0x2a),bpb[0x2a],bpb[0x2b]);
    printf("2ch-2fh Root Cluster Number: 0x%x %u\n",*(unsigned *)(bpb + 0x2c),*(unsigned *)(bpb + 0x2c));
    printf("30h-31h File System Information Sector Number: 0x%x %u\n",*(short *)(bpb + 0x30),*(short *)(bpb + 0x30));
    printf("32h-33h Backup Boot Sector : 0x%x %u\n",*(short *)(bpb + 0x32),*(short *)(bpb + 0x32));
    printf("34h-3fh Reserved: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",bpb[0x34],bpb[0x35],bpb[0x36],bpb[0x37],bpb[0x38],bpb[0x39],bpb[0x3a],bpb[0x3b],bpb[0x3c],bpb[0x3d],bpb[0x3e],bpb[0x3f]);
    printf("Extended BIOS Parameter Block\n");
    printf("40h     Physical Drive Number: 0x%x %d\n",bpb[0x40],bpb[0x40]);
    printf("41h     Reserved: 0x%x %d\n",bpb[0x41],bpb[0x41]);
    printf("42h     Extended Boot Signature: 0x%x %d\n",bpb[0x42],bpb[0x42]);
    printf("43h-46h Volume Serial Number: 0x%08x %u\n",*(unsigned *)(bpb + 0x43),*(unsigned *)(bpb + 0x43));
    printf("47h-51h Volume Label: \"%.11s\"\n",bpb + 0x47);
    printf("52h-59h System ID: \"%.8s\"\n",bpb + 0x52);
}

void fat(unsigned char *bpb)
{
    printf("00h-02h Jump Instruction: 0x%02x 0x%02x 0x%02x\n",bpb[0],bpb[1],bpb[2]);
    printf("03h-0ah OEM ID: \"%.8s\"\n",bpb + 0x3);
    printf("BIOS Parameter Block\n");
    printf("0bh-0ch Bytes Per Sector: 0x%x %d\n",*(short *)(bpb + 0x0b),*(short *)(bpb + 0x0b));
    printf("0dh     Sectors Per Cluster: 0x%x %d\n",bpb[0x0d],bpb[0x0d]);
    printf("0eh-0fh Reserved Sectors: 0x%x %d\n",*(short *)(bpb + 0x0e),*(short *)(bpb + 0x0e));
    printf("10h     Number of FATs: 0x%x %d\n",bpb[0x10],bpb[0x10]);
    printf("11h-12h Root Entries: 0x%x %u\n",*(unsigned short *)(bpb + 0x11),*(unsigned short *)(bpb + 0x11));
    printf("13h-14h Small Sectors: 0x%x %d\n",*(short *)(bpb + 0x13),*(short *)(bpb + 0x13));
    printf("15h     Media Descriptor: 0x%x %d\n",bpb[0x15],bpb[0x15]);
    printf("16h-17h Sectors Per FAT: 0x%x %d\n",*(short *)(bpb + 0x16),*(short *)(bpb + 0x16));
    printf("18h-19h Sectors Per Track: 0x%x %u\n",*(unsigned short *)(bpb + 0x18),*(unsigned short *)(bpb + 0x18));
    printf("1ah-1bh Number Of Heads: 0x%x %d\n",*(short *)(bpb + 0x1a),*(short *)(bpb + 0x1a));
    printf("1ch-1fh Hidden Sectors: 0x%x %u\n",*(unsigned *)(bpb + 0x1c),*(unsigned *)(bpb + 0x1c));
    printf("20h-23h Large Sectors: 0x%x %u\n",*(unsigned *)(bpb + 0x20),*(unsigned *)(bpb + 0x20));
    printf("Extended BIOS Parameter Block\n");
    printf("24h     Physical Drive Number: 0x%x %d\n",bpb[0x24],bpb[0x24]);
    printf("25h     Reserved: 0x%x %d\n",bpb[0x25],bpb[0x25]);
    printf("26h     Extended Boot Signature: 0x%x %d\n",bpb[0x26],bpb[0x26]);
    printf("27h-2ah Volume Serial Number: 0x%08x %u\n",*(unsigned *)(bpb + 0x27),*(unsigned *)(bpb + 0x27));
    printf("2bh-35h Volume Label: \"%.11s\"\n",bpb + 0x2b);
    printf("36h-3dh File System Type: \"%.8s\"\n",bpb + 0x36);
}

int main(int argc,char **argv)
{
    HANDLE hDevice;               // handle to the drive to be examined 
    int lasterror1;
    BOOL bResult;                   // generic results flag
    BOOL bResult1;                   // generic results flag
    char dev[512]="C:\\";
    char dev2[512]="\\\\.\\C:";
    UINT DriveType;
    char VolumeNameBuffer[BUFSIZE];
    DWORD VolumeSerialNumber;
    DWORD MaximumComponentLength;
    DWORD FileSystemFlags;
    char FileSystemNameBuffer[BUFSIZE];
    char diskbuf[512*8];
    DWORD NumberOfBytesRead;
    PARTITION_INFORMATION Partition_Infomation;
    DWORD BytesReturned;

    if ( argc >= 2 ) {
        dev[0] = argv[1][0];
        dev2[4] = dev[0];
    }

    DriveType = GetDriveType(dev);
    if ( DriveType != DRIVE_REMOVABLE && DriveType != DRIVE_FIXED && DriveType != DRIVE_RAMDISK ) {
        printf("DriveType %s not suported.\n",drivetypestring(DriveType));
        return 1;
    }
    bResult = GetVolumeInformation(
              dev,
              VolumeNameBuffer,
              sizeof VolumeNameBuffer,
              &VolumeSerialNumber,
              &MaximumComponentLength,
              &FileSystemFlags,
              FileSystemNameBuffer,
              sizeof FileSystemNameBuffer
            );
    if ( bResult == 0 ) 
    {
        lasterr("GetVolumeInformation(%s): ", dev);
        return 1;
    }
    hDevice = CreateFile(dev2,  // drive to open
                    GENERIC_READ ,    // 
                    FILE_SHARE_READ | // share mode
                    FILE_SHARE_WRITE, 
                    NULL,             // default security attributes
                    OPEN_EXISTING,    // disposition
                    FILE_FLAG_SEQUENTIAL_SCAN,   // file attributes
                    NULL);            // do not copy file attributes
    if (hDevice == INVALID_HANDLE_VALUE) // cannot open the drive
    {
        lasterr("CreateFile(%s): ", dev2);
        return 1;
    } else {
        bResult1 = DeviceIoControl(
                     hDevice,                // handle to a partition
                     IOCTL_DISK_GET_PARTITION_INFO,   // dwIoControlCode
                     NULL,                            // lpInBuffer
                     0,                               // nInBufferSize
                     &Partition_Infomation,            // output buffer
                     sizeof Partition_Infomation,  // size of output buffer
                     &BytesReturned,       // number of bytes returned
                     NULL      // OVERLAPPED structure
                   );
        lasterror1 = GetLastError();
        bResult = ReadFile(hDevice,diskbuf,sizeof diskbuf,&NumberOfBytesRead,NULL);
        if ( bResult == 0 ) 
        {
            lasterr("ReadFile(%s): ", dev2);
            return 1;
        }
        CloseHandle(hDevice);
    }

    printf("Volume: %s\n",dev);
    if ( bResult1 ) 
    {
        printf("StartingOffset: 0x%I64x %I64d\n",Partition_Infomation.StartingOffset.QuadPart,Partition_Infomation.StartingOffset.QuadPart);
        printf("PartitionLength: 0x%I64x %I64d (%.1fKB) (%gMB) (%gGB)\n",Partition_Infomation.PartitionLength.QuadPart,Partition_Infomation.PartitionLength.QuadPart,Partition_Infomation.PartitionLength.QuadPart / 1024.0,Partition_Infomation.PartitionLength.QuadPart / (1024.0 * 1024),Partition_Infomation.PartitionLength.QuadPart / (1024.0 * 1024 * 1024));
        printf("HiddenSectors: 0x%x %d\n",Partition_Infomation.HiddenSectors,Partition_Infomation.HiddenSectors);
        printf("PartitionNumber: 0x%x %d\n",Partition_Infomation.PartitionNumber,Partition_Infomation.PartitionNumber);
        printf("PartitionType: 0x%02x %d (%s)\n",Partition_Infomation.PartitionType,Partition_Infomation.PartitionType,partitionstring(Partition_Infomation.PartitionType));
        printf("BootIndicator: %s\n",Partition_Infomation.BootIndicator ? "TRUE" : "FALSE");
        printf("RecognizedPartition: %s\n",Partition_Infomation.RecognizedPartition ? "TRUE" : "FALSE");
        printf("RewritePartition: %s\n",Partition_Infomation.RewritePartition ? "TRUE" : "FALSE");
    }
    if ( strcmp(FileSystemNameBuffer,"NTFS") == 0 ) {
        ntfs(diskbuf);
    } else if ( strcmp(FileSystemNameBuffer,"FAT32") == 0 ) {
        fat32(diskbuf);
    } else if ( memcmp(FileSystemNameBuffer,"FAT",3) == 0 ) {
        fat(diskbuf);
    }
    return 0;
}

